//! A Linux implementation of the traits from `drivers::tungeneric`

use std::error::Error;
use std::io;
use std::io::{Read, Write};
use etherparse::IpHeader;
use tun::platform::Device;
use crate::drivers::tungeneric::{GenericDriver, GenericInterface, TunPacket};

#[allow(clippy::module_name_repetitions)]
/// Represents the internal state of a tun driver.
pub struct TunDevice {
    /// Contains the device handle of the tun device
    device: Device,
    /// A packet read buffer
    read_buf: [u8; 4096],
    /// Currently unused, for storing the offset into the read buffer to write to
    read_offset: usize,
    /// Another packet read buffer
    packet_buf: [u8; 4096],
}
impl GenericDriver for TunDevice {
    fn new(config: &GenericInterface) -> Result<Self, Box<dyn Error>> {
        let mut device_config = tun::Configuration::default();

        for address in &config.addresses {
            device_config.address(address.address()).netmask(address.mask());
        }

        device_config.name(&config.name);
        if let Some(mtu) = config.mtu {
            device_config.mtu(mtu);
        }
        Ok(Self {
            device: tun::create(&device_config)?,
            read_buf: [0u8; 4096],
            read_offset: 0,
            packet_buf: [0u8; 4096],
        })
    }

    fn read(&mut self) -> Result<TunPacket, Box<dyn Error>> {
        self.packet_buf = [0u8; 4096];
        let _ = self.device.read(&mut self.packet_buf)?;
        let ip_header = IpHeader::from_slice(&self.packet_buf);
        if let Ok((packet_header, _, _)) = ip_header {
            // Found a packet. Clear read buffers
            self.read_offset = 0;
            self.read_buf = [0u8; 4096];
            return Ok(TunPacket {
                header: packet_header,
                packet: self.packet_buf,
            })
        }

        Err(io::Error::new(io::ErrorKind::WouldBlock, "No packets detected yet").into())
    }

    fn clear(&mut self) {
        self.read_buf = [0u8; 4096];
        self.read_offset = 0;
    }

    fn write(&mut self, packet: TunPacket) -> Result<(), Box<dyn Error>> {
        let _ = self.device.write(&packet.packet)?;
        Ok(())
    }
}