//! Generic traits and modules to represent a tun/tap device

use std::error::Error;
use cidr::IpInet;
use etherparse::{IpHeader};

#[derive(Debug)]
/// A parsed packet from an interface
pub struct TunPacket {
    /// The parsed IPv4 or IPv6 header
    pub header: IpHeader,
    /// The raw packet data
    pub packet: [u8; 4096]
}

/// A trait to implement R/W on a tun driver
pub trait GenericDriver {
    /// Create a new device with the provided generic interface configuration
    /// # Errors
    /// This function may error if there is an error creating the device.
    fn new(config: &GenericInterface) -> Result<Self, Box<dyn Error>> where Self: Sized;

    /// Attempt to read a packet from the interface. If no packet is found, append it to an internal buffer and return `WouldBlock`.
    /// Upon finding a successful packet inside the internal buffer, or the .clear() method being called, the internal buffer will be cleared.
    /// # Errors
    /// This function will return `WouldBlock` if a packet could not be read. This function may error if there is an error reading the interface.
    fn read(&mut self) -> Result<TunPacket, Box<dyn Error>>;

    /// Clear any internal state of this interface.
    fn clear(&mut self);

    /// Attempt to write a packet to the interface.
    /// # Errors
    /// This function will error if an error occured writing to the device.
    fn write(&mut self, packet: TunPacket) -> Result<(), Box<dyn Error>>;
}

/// Generic interface config
pub struct GenericInterface {
    /// The ipv4/ipv6 addresses to be assigned to the interface
    pub addresses: Vec<IpInet>,
    /// The mtu to be assigned to the interface
    pub mtu: Option<i32>,
    /// The name of the virtual interface
    pub name: String
}