//! Various functions for ChaCha20Poly1305 stream ciphers

use chacha20poly1305::{ChaCha20Poly1305, Error, KeyInit, XChaCha20Poly1305};
use chacha20poly1305::aead::{Aead, Payload, Nonce};

/// Encrypt the plaintext with the given parameters using ChaCha20Poly1305
/// # Errors
/// This function will error if the encryption was unsuccessful
/// # Panics
/// This function, while having an .unwrap() call, will never panic because the key is a fixed, correctly sized array.
#[allow(clippy::module_name_repetitions)]
pub fn qcrypto_aead(key: &[u8; 32], counter: u64, plaintext: &[u8], authtext: &[u8]) -> Result<Vec<u8>, Error> {
    let cipher = ChaCha20Poly1305::new_from_slice(key).unwrap();
    let mut nonce_bytes = [0u8; 12];
    nonce_bytes[4..].copy_from_slice(&counter.to_le_bytes());
    let mut payload = Payload::from(plaintext);
    payload.aad = authtext;

    cipher.encrypt(&Nonce::<ChaCha20Poly1305>::from(nonce_bytes), payload)
}

/// Decrypt the ciphertext with the given parameters using ChaCha20Poly1305
/// # Errors
/// This function will error if the decryption was unsuccessful
/// # Panics
/// This function, while having an .unwrap() call, will never panic because the key is a fixed, correctly sized array.
pub fn qcrypto_aead_decrypt(key: &[u8; 32], counter: u64, ciphertext: &[u8], authtext: &[u8]) -> Result<Vec<u8>, Error> {
    let cipher = ChaCha20Poly1305::new_from_slice(key).unwrap();
    let mut nonce_bytes = [0u8; 12];
    nonce_bytes[4..].copy_from_slice(&counter.to_le_bytes());
    let mut payload = Payload::from(ciphertext);
    payload.aad = authtext;

    cipher.decrypt(&Nonce::<ChaCha20Poly1305>::from(nonce_bytes), payload)
}

/// Encrypt the ciphertext with the given parameters using XChaCha20Poly1305
/// # Errors
/// This function will error if the decryption was unsuccessful
/// # Panics
/// This function, while having an .unwrap() call, will never panic because the key is a fixed, correctly sized array.
pub fn qcrypto_xaead(key: &[u8; 32], nonce: &[u8; 24], plaintext: &[u8], authtext: &[u8]) -> Result<Vec<u8>, Error> {
    let cipher = XChaCha20Poly1305::new_from_slice(key).unwrap();
    let mut payload = Payload::from(plaintext);
    payload.aad = authtext;

    cipher.encrypt(Nonce::<XChaCha20Poly1305>::from_slice(nonce), payload)
}

/// Decrypt the ciphertext with the given parameters using XChaCha20Poly1305
/// # Errors
/// This function will error if the decryption was unsuccessful
/// # Panics
/// This function, while having an .unwrap() call, will never panic because the key is a fixed, correctly sized array.
pub fn qcrypto_xaead_decrypt(key: &[u8; 32], nonce: &[u8; 24], ciphertext: &[u8], authtext: &[u8]) -> Result<Vec<u8>, Error> {
    let cipher = XChaCha20Poly1305::new_from_slice(key).unwrap();
    let mut payload = Payload::from(ciphertext);
    payload.aad = authtext;

    cipher.decrypt(Nonce::<XChaCha20Poly1305>::from_slice(nonce), payload)
}