//! Various hash functions

use blake2::{Blake2s256, Blake2sMac, Digest};
use blake2::digest::{KeyInit, FixedOutput, Update};
use hmac::SimpleHmac;

type HmacBlake2s = SimpleHmac<Blake2s256>;

/// Given a varied length input, produce a 32-byte Blake2s hash digest
pub fn qcrypto_hash(input: &[u8]) -> [u8; 32] {
    let mut hasher = Blake2s256::new();
    Update::update(&mut hasher, input);
    hasher.finalize().into()
}

/// Given two varied length inputs, produce a 32-byte Blake2s hash digest
pub fn qcrypto_hash_twice(input: &[u8], input2: &[u8]) -> [u8; 32] {
    let mut hasher = Blake2s256::new();
    Update::update(&mut hasher, input);
    Update::update(&mut hasher, input2);
    hasher.finalize().into()
}

/// Given a varied length MAC key and a varied length input, produce a 16-byte MAC digest using Blake2s
/// # Panics
/// This function will panic if the key is an incorrect size.
pub fn qcrypto_mac(key: &[u8], input: &[u8]) -> [u8; 16] {
    let mut hasher = Blake2sMac::new_from_slice(key).unwrap();
    hasher.update(input);
    hasher.finalize_fixed().into()
}

/// Given a varied length HMAC key and a varied length input, produce a 32-byte HMAC digest using Blake2s
/// # Panics
/// This function will panic if the key is an incorrect size.
pub fn qcrypto_hmac(key: &[u8], input: &[u8]) -> [u8; 32] {
    let mut hasher = HmacBlake2s::new_from_slice(key).unwrap();
    Update::update(&mut hasher, input);
    hasher.finalize_fixed().into()
}

/// Given a varied length HMAC key and two varied length inputs, produce a 32-byte HMAC digest using Blake2s.
/// This does essentially the same thing as concatenating input2 to input and calling `qcrypto_hmac` on that.
/// # Panics
/// This function will panic if the key is an incorrect size.
pub fn qcrypto_hmac_twice(key: &[u8], input: &[u8], input2: &[u8]) -> [u8; 32] {
    let mut hasher = HmacBlake2s::new_from_slice(key).unwrap();
    Update::update(&mut hasher, input);
    Update::update(&mut hasher, input2);
    hasher.finalize_fixed().into()
}